@@ -3,13 +3,13 @@ from abc import abstractmethod
 from collections import defaultdict
 from collections import deque
 from collections import namedtuple
+from contextlib import ExitStack
 from datetime import datetime
 from typing import DefaultDict
 from typing import Deque
 from typing import Dict
 from typing import Iterable
 from typing import Iterator
-from typing import Literal
 from typing import Optional
 from typing import Sequence
 from typing import Set
@@ -46,12 +46,12 @@ from dipdup.exceptions import ReindexingReason
 from dipdup.models import BigMapAction
 from dipdup.models import BigMapData
 from dipdup.models import BigMapDiff
-from dipdup.models import BlockData
 from dipdup.models import HeadBlockData
 from dipdup.models import IndexStatus
 from dipdup.models import OperationData
 from dipdup.models import Origination
 from dipdup.models import Transaction
+from dipdup.prometheus import Metrics
 from dipdup.utils import FormattedLogger
 from dipdup.utils.database import in_global_transaction
 
@@ -67,10 +67,6 @@ class OperationSubgroup:
     operations: Tuple[OperationData, ...]
     entrypoints: Set[Optional[str]]
 
-    @property
-    def length(self) -> int:
-        return len(self.operations)
-
     def __hash__(self) -> int:
         return hash(f'{self.hash}:{self.counter}')
 
@@ -83,10 +79,6 @@ OperationHandlerArgumentT = Optional[Union[Transaction, Origination, OperationDa
 MatchedOperationsT = Tuple[OperationSubgroup, OperationHandlerConfig, Deque[OperationHandlerArgumentT]]
 MatchedBigMapsT = Tuple[BigMapHandlerConfig, BigMapDiff]
 
-# NOTE: For initializing the index state on startup
-block_cache: Dict[Tuple[str, int], BlockData] = {}
-head_cache: DefaultDict[str, Optional[Union[models.Head, Literal[False]]]] = defaultdict(lambda: False)
-
 
 def extract_operation_subgroups(
     operations: Iterable[OperationData],
@@ -161,6 +153,18 @@ class Index:
             raise RuntimeError('Index state is not initialized')
         return self._state
 
+    @property
+    def queue_size(self) -> int:
+        return len(self._queue)
+
+    @property
+    def synchronized(self) -> bool:
+        return self.state.status == IndexStatus.REALTIME
+
+    @property
+    def realtime(self) -> bool:
+        return self.state.status == IndexStatus.REALTIME and not self._queue
+
     async def initialize_state(self, state: Optional[models.Index] = None) -> None:
         if self._state:
             raise RuntimeError('Index state is already initialized')
@@ -188,7 +192,10 @@ class Index:
         # NOTE: `--oneshot` flag implied
         if isinstance(self._config, (OperationIndexConfig, BigMapIndexConfig)) and self._config.last_level:
             last_level = self._config.last_level
-            await self._synchronize(last_level, cache=True)
+            with ExitStack() as stack:
+                if Metrics.enabled:
+                    stack.enter_context(Metrics.measure_total_sync_duration())
+                await self._synchronize(last_level, cache=True)
             await self.state.update_status(IndexStatus.ONESHOT, last_level)
 
         sync_levels = {self.datasource.get_sync_level(s) for s in self._config.subscriptions}
@@ -203,10 +210,17 @@ class Index:
         elif level < sync_level:
             self._logger.info('Index is behind datasource, sync to datasource level: %s -> %s', level, sync_level)
             self._queue.clear()
-            await self._synchronize(sync_level)
 
-        else:
-            await self._process_queue()
+            with ExitStack() as stack:
+                if Metrics.enabled:
+                    stack.enter_context(Metrics.measure_total_sync_duration())
+                await self._synchronize(sync_level)
+
+        elif self._queue:
+            with ExitStack() as stack:
+                if Metrics.enabled:
+                    stack.enter_context(Metrics.measure_total_realtime_duration())
+                await self._process_queue()
 
     @abstractmethod
     async def _synchronize(self, last_level: int, cache: bool = False) -> None:
@@ -233,6 +247,8 @@ class Index:
 
     async def _exit_sync_state(self, last_level: int) -> None:
         self._logger.info('Index is synchronized to level %s', last_level)
+        if Metrics.enabled:
+            Metrics.set_levels_to_sync(self._config.name, 0)
         await self.state.update_status(status=IndexStatus.REALTIME, level=last_level)
 
     def _extract_level(self, message: Union[Tuple[OperationData, ...], Tuple[BigMapData, ...]]) -> int:
@@ -250,11 +266,13 @@ class OperationIndex(Index):
         self._queue: Deque[OperationQueueItemT] = deque()
         self._contract_hashes: Dict[str, Tuple[int, int]] = {}
         self._rollback_level: Optional[int] = None
-        self._head_hashes: Set[str] = set()
+        self._head_hashes: Dict[str, bool] = {}
         self._migration_originations: Optional[Dict[str, OperationData]] = None
 
     def push_operations(self, operation_subgroups: Tuple[OperationSubgroup, ...]) -> None:
         self._queue.append(operation_subgroups)
+        if Metrics.enabled:
+            Metrics.set_levels_to_realtime(self._config.name, len(self._queue))
 
     def push_rollback(self, level: int) -> None:
         self._queue.append(SingleLevelRollback(level))
@@ -280,12 +298,20 @@ class OperationIndex(Index):
         """Process WebSocket queue"""
         while self._queue:
             message = self._queue.popleft()
+            messages_left = len(self._queue)
+            if Metrics.enabled:
+                Metrics.set_levels_to_realtime(self._config.name, messages_left)
             if isinstance(message, SingleLevelRollback):
-                self._logger.debug('Processing rollback realtime message, %s left in queue', len(self._queue))
+                self._logger.debug('Processing rollback realtime message, %s left in queue', messages_left)
                 await self._single_level_rollback(message.level)
             elif message:
-                self._logger.debug('Processing operations realtime message, %s left in queue', len(self._queue))
-                await self._process_level_operations(message)
+                self._logger.debug('Processing operations realtime message, %s left in queue', messages_left)
+                with ExitStack() as stack:
+                    if Metrics.enabled:
+                        stack.enter_context(Metrics.measure_level_realtime_duration())
+                    await self._process_level_operations(message)
+        else:
+            Metrics.set_levels_to_realtime(self._config.name, 0)
 
     async def _synchronize(self, last_level: int, cache: bool = False) -> None:
         """Fetch operations via Fetcher and pass to message callback"""
@@ -315,6 +341,9 @@ class OperationIndex(Index):
         )
 
         async for level, operations in fetcher.fetch_operations_by_level():
+            if Metrics.enabled:
+                Metrics.set_levels_to_sync(self._config.name, last_level - level)
+
             operation_subgroups = tuple(
                 extract_operation_subgroups(
                     operations,
@@ -324,7 +353,10 @@ class OperationIndex(Index):
             )
             if operation_subgroups:
                 self._logger.info('Processing operations of level %s', level)
-                await self._process_level_operations(operation_subgroups)
+                with ExitStack() as stack:
+                    if Metrics.enabled:
+                        stack.enter_context(Metrics.measure_level_sync_duration())
+                    await self._process_level_operations(operation_subgroups)
 
         await self._exit_sync_state(last_level)
 
@@ -343,21 +375,29 @@ class OperationIndex(Index):
                 levels_repr = ', '.join(f'{k}={v}' for k, v in levels.items())
                 raise RuntimeError(f'Index is in a rollback state, but received operation batch with different levels: {levels_repr}')
 
-            self._logger.info('Rolling back to previous level, verifying processed operations')
-            received_hashes = {s.hash for s in operation_subgroups}
-            new_hashes = received_hashes - self._head_hashes
-            missing_hashes = self._head_hashes - received_hashes
+            self._logger.info('Rolling back to the previous level, verifying processed operations')
+            old_head_hashes = set(self._head_hashes)
+            old_head_matched_hashes = {k for k, v in self._head_hashes.items() if v}
+            new_head_hashes = {s.hash for s in operation_subgroups}
+            unprocessed_hashes = new_head_hashes - old_head_hashes
+            # NOTE: We can ignore subgroups that don't match any handlers
+            missing_hashes = old_head_matched_hashes - new_head_hashes
 
-            self._logger.info('Comparing hashes: %s new, %s missing', len(new_hashes), len(missing_hashes))
+            self._logger.info('Comparing hashes: %s new, %s missing', len(unprocessed_hashes), len(missing_hashes))
             if missing_hashes:
-                self._logger.info('Some operations were backtracked: %s', ', '.join(missing_hashes))
-                # TODO: More context
-                await self._ctx.reindex(ReindexingReason.ROLLBACK)
+                self._logger.info('Some operations were backtracked, requesting reindexing')
+                await self._ctx.reindex(
+                    ReindexingReason.rollback,
+                    datasource=self._datasource.name,
+                    from_level=level + 1,
+                    to_level=level,
+                    missing_hashes=', '.join(missing_hashes),
+                )
 
             self._rollback_level = None
             self._head_hashes.clear()
 
-            operation_subgroups = tuple(filter(lambda subgroup: subgroup.hash in new_hashes, operation_subgroups))
+            operation_subgroups = tuple(filter(lambda subgroup: subgroup.hash in unprocessed_hashes, operation_subgroups))
 
         elif level < self.state.level:
             raise RuntimeError(f'Level of operation batch must be higher than index state level: {level} < {self.state.level}')
@@ -365,8 +405,12 @@ class OperationIndex(Index):
         self._logger.debug('Processing %s operation subgroups of level %s', len(operation_subgroups), level)
         matched_handlers: Deque[MatchedOperationsT] = deque()
         for operation_subgroup in operation_subgroups:
-            self._head_hashes.add(operation_subgroup.hash)
-            matched_handlers += await self._match_operation_subgroup(operation_subgroup)
+            subgroup_matched_handlers = await self._match_operation_subgroup(operation_subgroup)
+            matched_handlers += subgroup_matched_handlers
+            self._head_hashes[operation_subgroup.hash] = bool(subgroup_matched_handlers)
+
+        if Metrics.enabled:
+            Metrics.set_index_handlers_matched(len(matched_handlers))
 
         # NOTE: We still need to bump index level but don't care if it will be done in existing transaction
         if not matched_handlers:
@@ -564,13 +608,19 @@ class BigMapIndex(Index):
     def push_big_maps(self, big_maps: Tuple[BigMapData, ...]) -> None:
         self._queue.append(big_maps)
 
+        if Metrics.enabled:
+            Metrics.set_levels_to_realtime(self._config.name, len(self._queue))
+
     async def _process_queue(self) -> None:
         """Process WebSocket queue"""
         if self._queue:
             self._logger.debug('Processing websocket queue')
         while self._queue:
             big_maps = self._queue.popleft()
-            await self._process_level_big_maps(big_maps)
+            with ExitStack() as stack:
+                if Metrics.enabled:
+                    stack.enter_context(Metrics.measure_level_realtime_duration())
+                await self._process_level_big_maps(big_maps)
 
     async def _synchronize(self, last_level: int, cache: bool = False) -> None:
         """Fetch operations via Fetcher and pass to message callback"""
@@ -605,8 +655,12 @@ class BigMapIndex(Index):
             cache=cache,
         )
 
-        async for _, big_maps in fetcher.fetch_big_maps_by_level():
-            await self._process_level_big_maps(big_maps)
+        async for level, big_maps in fetcher.fetch_big_maps_by_level():
+            with ExitStack() as stack:
+                if Metrics.enabled:
+                    Metrics.set_levels_to_sync(self._config.name, last_level - level)
+                    stack.enter_context(Metrics.measure_level_sync_duration())
+                await self._process_level_big_maps(big_maps)
 
     async def _synchronize_level(self, last_level: int, cache: bool = False) -> None:
         # NOTE: Checking late because feature flags could be modified after loading config
@@ -617,35 +671,38 @@ class BigMapIndex(Index):
         big_map_ids: Set[Tuple[int, str, str]] = set()
 
         for address, path in big_map_pairs:
-            for contract_big_map in await self._datasource.get_contract_big_maps(address):
-                if contract_big_map['path'] == path:
-                    big_map_ids.add((int(contract_big_map['ptr']), address, path))
+            async for contract_big_maps in self._datasource.iter_contract_big_maps(address):
+                for contract_big_map in contract_big_maps:
+                    if contract_big_map['path'] == path:
+                        big_map_ids.add((int(contract_big_map['ptr']), address, path))
 
         # NOTE: Do not use `_process_level_big_maps` here; we want to maintain transaction manually.
         async with in_global_transaction():
             for big_map_id, address, path in big_map_ids:
-                big_map_keys = await self._datasource.get_big_map(big_map_id, last_level)
-                big_map_data = tuple(
-                    BigMapData(
-                        id=big_map_key['id'],
-                        level=last_level,
-                        operation_id=last_level,
-                        timestamp=datetime.now(),
-                        bigmap=big_map_id,
-                        contract_address=address,
-                        path=path,
-                        action=BigMapAction.ADD_KEY,
-                        active=big_map_key['active'],
-                        key=big_map_key['key'],
-                        value=big_map_key['value'],
+                async for big_map_keys in self._datasource.iter_big_map(big_map_id, last_level):
+                    big_map_data = tuple(
+                        BigMapData(
+                            id=big_map_key['id'],
+                            level=last_level,
+                            operation_id=last_level,
+                            timestamp=datetime.now(),
+                            bigmap=big_map_id,
+                            contract_address=address,
+                            path=path,
+                            action=BigMapAction.ADD_KEY,
+                            active=big_map_key['active'],
+                            key=big_map_key['key'],
+                            value=big_map_key['value'],
+                        )
+                        for big_map_key in big_map_keys
                     )
-                    for big_map_key in big_map_keys
-                )
-                matched_handlers = await self._match_big_maps(big_map_data)
-                for handler_config, big_map_diff in matched_handlers:
-                    await self._call_matched_handler(handler_config, big_map_diff)
+                    matched_handlers = await self._match_big_maps(big_map_data)
+                    for handler_config, big_map_diff in matched_handlers:
+                        await self._call_matched_handler(handler_config, big_map_diff)
 
-    async def _process_level_big_maps(self, big_maps: Tuple[BigMapData, ...]):
+            await self.state.update_status(level=last_level)
+
+    async def _process_level_big_maps(self, big_maps: Tuple[BigMapData, ...]) -> None:
         if not big_maps:
             return
         level = self._extract_level(big_maps)
